static char *rcsid_hyph_c="$Id: hyph.c,v 1.2 1992/07/16 21:42:50 rosenkra Exp $";

/*
 * $Log: hyph.c,v $
 * Revision 1.2  1992/07/16  21:42:50  rosenkra
 * new at this rev. ported to gcc.
 *
 */

#undef DEBUG_HYPHEN
#undef STANDALONE

#ifdef STANDALONE
# ifndef DEBUG_HYPHEN
#  define DEBUG_HYPHEN
# endif
#endif

/*
 *	hyph - An implementation of the Knuth hyphenation algorithm, modified
 *
 *	currently, hyphen ret 0 if cannot be hyphenated, 1 for it was (or is)
 *	soft hyphenated. should modify so that ret -1 if soft hyphenated or
 *	n (positive) where n is index into string where there is a hard hyphen.
 *	either that or check for a hard hyphen before even calling this routine
 *
 *	two entry points here:
 *
 *	char *do_hyphen (char *, int, char **, char **);
 *	int hyphen (char *, char *);
 */

#include <stdio.h>
#include <ctype.h>

#ifdef ALCYON
# ifndef void
#  define void		int
# endif
#endif

/*
 *	#defines for non-printing ASCII characters
 */
#define	NUL		0x00	/* ^@	*/
#define	SOH		0x01	/* ^A	*/
#define	STX		0x02	/* ^B	*/
#define	ETX		0x03	/* ^C	*/
#define	EOT		0x04	/* ^D	*/
#define	ENQ		0x05	/* ^E	*/
#define	ACK		0x06	/* ^F	*/
#define	BEL		0x07	/* ^G	*/
#define	BS		0x08	/* ^H	*/
#define	HT		0x09	/* ^I	*/
#define	LF		0x0a	/* ^J	*/
#define NL		LF
#define	VT		0x0b	/* ^K	*/
#define	FF		0x0c	/* ^L	*/
#define	CR		0x0d	/* ^M	*/
#define	SO		0x0e	/* ^N	*/
#define	SI		0x0f	/* ^O	*/
#define	DLE		0x10	/* ^P	*/
#define	DC1		0x11	/* ^Q	*/
#define	DC2		0x12	/* ^R	*/
#define	DC3		0x13	/* ^S	*/
#define	DC4		0x14	/* ^T	*/
#define	NAK		0x15	/* ^U	*/
#define	SYN		0x16	/* ^V	*/
#define	ETB		0x17	/* ^W	*/
#define	CAN		0x18	/* ^X	*/
#define	EM		0x19	/* ^Y	*/
#define	SUB		0x1a	/* ^Z	*/
#define CPMEOF		SUB
#define	ESC		0x1b	/* ^[	*/
#define	FS		0x1c	/* ^\	*/
#define	GS		0x1d	/* ^]	*/
#define	RS		0x1e	/* ^^	*/
#define	US		0x1f	/* ^_	*/
#define	DEL		0x7f	/* DEL	*/





/*
 *	if THING is #defined then words like "anything" will be split
 *	any-thing. On the other hand, if it's defined the bathing is
 *	split ba-thing. You can't have every-thing. In the present
 *	version most of the -thing words are in the exception list.
 *
 * #define THING
 */

/*
 *	Various psuedo-subroutines. HYPHEN defines a bit to set when a hyphen
 *	is inserted. HYPHENATE sets the bit, UNHYPHENATE clears it,
 *	HAS_HYPHEN tests for it. The ER macro checks for an "er" at the end
 *	of a word, it's used by the consonant pair checking routine.
 *	Is consonant returns true if c is a consonant.
 *
 *	Note that a letter is marked if a hyphen can be inserted in front
 *	of it.
 */
#define HYPHEN		0x80 			/* soft hyphen bit */
#define HYPHENATE(c)	((c) |=  HYPHEN )	/* mark char */
#define UNHYPHENATE(c)	((c) &= ~HYPHEN )	/* unmark char */
#define HAS_HYPHEN(c)	((c) &   HYPHEN )	/* is it marked? */

#define ER(p,end)	((*p & 0x7f)=='e' && (*(p+1) & 0x7f)=='r' && (p+1)==end)

#define isvowel(c)	(islower((c) & 0x7f)  &&  vt[((c) & 0x7f) - 'a'])
#define isconsonant(c)	((c) && !isvowel(c))


/*	The dipthongs ch, gh, ph, sh and th are treated as single
 *	consonants. The subroutine nextch() will map these two
 *	characters into a single character as follows:
 */
#define CH		('z'+1)		/* {	 0x7b  \173    */
#define GH		('z'+2)		/* |	 0x7c  \174    */
#define PH		('z'+3)		/* }	 0x7d  \175    */
#define SH		('z'+4)		/* ~	 0x7e  \176    */
#define TH		('z'+5)		/* DEL	 0x7f  \177    */


/*
 *	global variables
 */
#ifdef DEBUG_HYPHEN
static int      HyDebug = 0;	/* True if debug diagnostics are to be printed*/
static void	bprint ();
#endif
static char	      **States;	/* Ptr to table for current state machine*/
static char		vt[] =
			{ 1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0 };
		       /* a b c d e f g h i j k l m n o p q r s t u v w x y z */

/*
 *	functions
 */
char	       *do_hyphen ();	/* global entries */

int		hyphen ();

static char    *suffix ();
static char    *prefix ();
static char	nextch ();
static int	isweird ();
static void	consonants ();	/* these are all local (static)... */
static void	phyphen ();
static int	next ();
static int	exception ();
static int	wcmp ();
static char    *bsearch ();

#ifdef __GNUC__
#include <string.h>
#else
extern char    *index ();	/* these are external... */
#endif


#define MAXWLEN		132




/*------------------------------*/
/*	do_hyphen		*/
/*------------------------------*/
char *do_hyphen (ps, n, p1, p2)
register char  *ps;		/* word to hyphenate */
int		n;		/* max number of chars before hyphen */
				/* if < 0, hyphenate anyway */
				/* if -1, hyphenate first avail location */
				/* if -2, hyphenate last avail location */
				/* if -3, hyphenate all locations */
char	      **p1;		/* -> first part up to and including hyphen */
char	      **p2;		/* -> last part after hyphen */
{

/*
 *	main entry point. space is allocated here so caller MUST use returned
 *	pointer before next call. ptr to hyphenated word is returned or NULL
 *	if it can't be hyphenated. word must be null terminated. max word len
 *	allowed is MAXWLEN.
 *
 *	use:
 *
 *	char *p1, *p2, *ps;
 *
 *	if (ps = do_hyphen ("hyphenate", 4, &p1, &p2))
 *	{
 *		deal with hyphen:
 *
 *		ps -> "hy-phenate"
 *		p1 -> "hy-..."
 *		p2 -> "phenate"
 *	}
 */

	static char		buf[MAXWLEN+2];		/* result */
	char			ibuf[MAXWLEN+2];	/* work space */
	register char	       *ebuf;			/* -> end of word */
	register char	       *p;
	register char	       *pb;
	register int		cnt;



	/*
	 *   too short? too long?
	 */
	if (strlen (ps) < 4 || strlen (ps) > MAXWLEN)
		return ((char *) 0);



	/*
	 *   copy word to input buffer
	 */
	for (ebuf = ibuf; *ps; ps++, ebuf++)
		*ebuf = *ps & 0x7f;
	*ebuf = '\0';



	/*
	 *   check for hard hyphen
	 */
	for (p = ibuf; *p; p++)
	{
		if (*p == '-')
		{
			if (n >= 0)
			{
				if (p - ibuf + 1 > n)
					return ((char *) 0);
			}
			strcpy (buf, ibuf);
			*p1 = buf;
			*p2 = (char *) (buf + (p - ibuf + 1));
			return (buf);
		}
	}



	/*
	 *   do it. check if it worked. if not, return null
	 */
	ebuf--;		/* -> last char */
	if (hyphen (ibuf, ebuf))
	{
		if (n >= 0)
		{
			/*
			 *   n was given. check if hyphen fits length
			 *   needed...
			 */
			for (cnt = 1, p = ibuf, pb = buf; *p && p <= ebuf;
			p++, pb++, cnt++)
			{
				if (HAS_HYPHEN(*p))
				{
					if (cnt > n)
						return ((char *) 0);
					else
					{
						*pb++ = '-';
						*p1 = buf;
						*p2 = pb;
						while (*p)
							*pb++ = *p++ & 0x7f;
						*pb = '\0';
						return (buf);
					}
				}
				*pb = *p & 0x7f;
				*(pb + 1) = '\0';
			}
			return ((char *) 0);
		}
		else if (n == -1)
		{
			/*
			 *   -1 was given. hyphenate first avail location
			 */
			for (p = ibuf, pb = buf; *p && p <= ebuf; p++, pb++)
			{
				if (HAS_HYPHEN(*p))
				{
					*pb++ = '-';
					*p1 = buf;
					*p2 = pb;
					while (*p)
						*pb++ = *p++ & 0x7f;
					*pb = '\0';
					return (buf);
				}
				*pb = *p & 0x7f;
				*(pb + 1) = '\0';
			}
			return ((char *) 0);
		}
		else if (n == -2)
		{
			/*
			 *   -2 was given. hyphenate last avail location
			 */
			pb = &buf[MAXWLEN];
			*pb = '\0';
			for (p = &ibuf[strlen(ibuf)-1], pb--;
			p >= ibuf; p--, pb--)
			{
				if (HAS_HYPHEN(*p))
				{
					*p1 = buf;
					*p2 = pb;
					*pb-- = '-';
					while (p >= ibuf)
						*pb-- = *p-- & 0x7f;
					pb++;
					return (pb);
				}
				*pb = *p & 0x7f;
				*(pb + 1) = '\0';
			}
			return (buf);
		}
		else if (n == -3)
		{
			/*
			 *   -3 was given. hyphenate all locations
			 */
			for (p = ibuf, pb = buf; *p && p <= ebuf; p++, pb++)
			{
				if (HAS_HYPHEN(*p))
				{
					*pb++ = '-';
					*p1 = buf;
					*p2 = pb;
				}
				*pb = *p & 0x7f;
				*(pb + 1) = '\0';
			}
			*pb = '\0';
			return (buf);
		}
		else
			return ((char *) 0);	/* error in n */
	}

	return ((char *) 0);
}




/*------------------------------*/
/*	hyphen			*/
/*------------------------------*/
int hyphen (beg, end)
char   *beg,
       *end;
{

/*
 *	main entry point
 *
 *	Hyphenate the word deliniated by beg and end: First
 *	strip suffixes, then strip a trailing s e or ed,
 *	then strip prefixes. Only words with more than four
 *	letters and which consist of lower case letters only
 *	(no NULLs) will be hyphenated. only 7-bit ascii chars are
 *	valid! (since soft hyphens are really the high bit set
 *	in chars which can have hyphens before them).
 *
 *	0 is returned if no attempt was made to hyphenate the word,
 *	1 otherwise.
 *	If the word has already been hyphenated, 1 is returned
 *	but the word is not modified.
 */

#ifdef DEBUG_HYPHEN
	extern int	Verbose;
#endif
	register char  *prefixp,
	               *suffixp;
	int		c;



	/*
	 *   too short?
	 */
	if (end - beg <= 4)
		return (0);


	/*
	 *   check for illegal conditions or if already hyphenated
	 */
	for (prefixp = beg; prefixp <= end; prefixp++)
	{
		/*
		 *   if already (soft) hyphenated, we are done...
		 */
		if (HAS_HYPHEN (*prefixp))
			return (1);

		/*
		 *   if not lower case, also done...
		 */
		if (!islower (*prefixp))
			return (0);
	}



	/*
	 *   check against exception list...
	 */
	if (exception (beg, end))
		return (1);



	/*
	 *   Hyphenate and remove any suffixes
	 */
	if ((suffixp = suffix (beg, end)) == end)
	{
		/*
		 *   If root has a trailing s e or ed remove it
		 *   before applying any other rules
		 */
		c = *end & 0x7f;

		if (c == 's' || c == 'e')
			suffixp = end - 1;

		else if ((*(end - 1) & 0x7f) == 'e' && c == 'd')
			suffixp = end - 2;
	}



	/*
	 *   Hyphenate and remove any prefixes
	 */
	prefixp = prefix (beg, suffixp);


#ifdef DEBUG_HYPHEN
	/*------------------------------------------------------*/
	/*  Print the word in the form:	 prefixes/root/suffixes */
	/* 							*/
	if (Verbose)
	{
		printf ( "\nprefix: <");
		bprint (beg, prefixp - 1);	/* Print prefixes*/
		printf (">\nroot:   <");
		bprint (prefixp, suffixp);	/* Print middle*/
		printf (">\nsuffix: <");
		bprint (suffixp + 1, end);	/* Print suffixes*/
		printf (">\n\n");
	}
	/*------------------------------------------------------*/
#endif


	if ((suffixp - prefixp) >= 3)
	{
		/*
		 *   Apply the consonant pairs rules only if the word
		 *   has at least 4 letters in it (after prefixes
		 *   and suffixes have been removed
		 */
		consonants (prefixp, suffixp);
	}

#ifdef DEBUG_HYPHEN
	if (Verbose)
		phyphen (beg, end);
#endif

	return (1);
}




/*------------------------------*/
/*	suffix			*/
/*------------------------------*/
static char *suffix (beg, end)
char   *beg,
       *end;
{

/*
 * 	Split off any suffixes, using the "Suffixes" state table
 *	(defined below) to recognize them.
 */

	extern char	      *Suffixes[];	/* Suffix state table */
	extern char	     **States;		/* Ptr to table for current state machine*/

	register char	       *p,
	                	c,
	                	c2;
	register unsigned int   state,
	                        times;

	state  = 1;
	times  = 0;
	States = Suffixes;

	for (p = end; p >= beg; )
	{
		c = *p-- & 0x7f;
		state = (unsigned int) next (state, c);

		if (!('a' <= c && c <= 'z'))	/*  Words containing non- */
			return (end);		/* lower-case characters */
						/* won't be hyphenated.	 */

		/*
		 *   Test for state 0 with an if instead of in the switch.
		 *   State 0 is weird; we use it to test for trailing e or
		 *   ed if we've failed. The first time through we set p to
		 *   end so we can check the end of the word. The third time
		 *   through we abort. This means that we'll retry if the
		 *   word ends -s -e -d -ed -de -dd -ee. The extra endings
		 *   shouldn't matter.
		 */
		if (state == 0)
		{
			if (times == 0)
				p = end;	/* Strip trailing e or d */

			else if (times == 1)	/* Strip trailing ed, de */
				p = end - 1;	/* ee, or dd	       */
			else
				return (end);

			times++;
		}

		switch (state)
		{
		case 86: 
			/*
			 *   Process -ing. This could have been done with
			 *   more states but it seemed easier to do it here.
			 *
			 *   I've modified the Knuth algorithm to not
			 *   hyphenate in front of a th that precedes
			 *   an -ing to make any-thing split right. Note
			 *   that bathing will split incorrectly though.
			 *   #define THING to enable this function.
			 *
			 *   On entry p points at the character preceding
			 *   the 'i'.
			 */
			c = *p & 0x7f;
			c2 = *(p-1) & 0x7f;

			if (p - beg < 3)
			{
				return (end);	/* FAIL */
			}
#ifdef THING
			else if (c == 'h' && c2 == 't')
			{
				p -= 2;
			}
#endif
			else if ((c == c2) && (!isvowel (c))
			&& (c != 'f') && (c != 's')
			&& (c != 'l') && (c != 'z'))
			{
				--p;
			}
			else if (c == 'l' && index ("bcdfghkptz", c2))
			{
				p -= (c2=='k' && (*(p-2) & 0x7f)=='c')
					? 1 : 2;
			}

			/* fall through to state 87 */

		case 87: 
			HYPHENATE (*(p+1));
			return (p);

		case 88: 
			HYPHENATE (*(p+2));
			return (p+1);

		case 89: 
			HYPHENATE (*(p+5));
			HYPHENATE (*(p+1));
			return (p);

		case 82: 
			p++;		/* p += 3 */
		case 81: 
			p++;		/* p += 2 */
		case 80: 
			p++;		/* P += 1 */
		case 83: 
			end = p;
			HYPHENATE (*(p+1));
			state = 1;
			break;

		case 84: 		/* FAILURE STATE */
			return (end);
		}
	}

	return (end);
}




/*------------------------------*/
/*	prefix			*/
/*------------------------------*/
static char *prefix (beg, end)
char   *beg,
       *end;
{

/*
 *	Split off any prefixes, using the "Prefixes" state table
 */

	extern char    *Prefixes[];	/* Prefix state table */
	extern char   **States;		/* Ptr to table for current state machine*/

	register char  *p,
			c;
	register int    state;

	state  = 1;
	States = Prefixes;

	for (p = beg; p < end; )
	{
		c     = (*p++ & 0x7f);
		state = next (state, c);

		switch (state)
		{
		case 82: 
			HYPHENATE (*p);
			HYPHENATE (*(p-1));
			return (p);
		case 83: 
			HYPHENATE (*p);
			HYPHENATE (*(p-2));
			return (p);
		case 84: 
			HYPHENATE (*p);
			HYPHENATE (*(p-3));
			return (p);

		case 81: 
			--p;
		case 87: 
			HYPHENATE (*p);
			return (p);

		case 85: 
			--p;
		case 86: 
			beg = p;
			state = 1;
		case 69: 
			HYPHENATE (*p);
			break;

		case 70: 
			HYPHENATE (*(p - 1));
			break;

		}
	}

	return (beg);			/* Failure */
}


/*
 *	This table is used to find exceptions to the VCCV rule. It is
 *	indexed by the first of the two consonents & contains pointers
 *	to strings, any character of which will form an exception if
 *	it's the second consonant.
 */

static char   *vccv_except[] =
{
 /* a  */	"",
 /* b  */	"lr",
 /* c  */	"lr",
 /* d  */	"gr",
 /* e  */	"",
 /* f  */	"lr",
 /* g  */	"lr",
 /* h  */	"",
 /* i  */	"",
 /* j  */	"",
 /* k  */	"n",
 /* l  */	"kq",
 /* m  */	"",
 /* n  */	"{kx",		/* CH, k, x */
 /* o  */	"",
 /* p  */	"lr",
 /* q  */	"",
 /* r  */	"k",
 /* s  */	"pq",
 /* t  */	"{r",		/* CH, r    */
 /* u  */	"",
 /* v  */	"",
 /* w  */	"hlnr",
 /* x  */	"",
 /* y  */	"",
 /* z  */	"",
 /* CH */	"lr",
 /* GH */	"t",
 /* PH */	"r",
 /* SH */	"",
 /* TH */	"r"
};




/*------------------------------*/
/*	nextch			*/
/*------------------------------*/
static char nextch (pp, endp)
char  **pp,
       *endp;
{

/*
 *	Advance *pp (one for a vowel or consonant, 2 for a dipthong)
 *	and return the character we've just skipped. Dipthongs are
 *	mapped to the single characters: TH, SH, PH, CH or GH.
 *	0 is returned once *pp reaches or passes endp.
 */

	register char	rval,
		       *p;

	if ((p = *pp) > endp)
		return ((char) 0);

	rval = *p++ & 0x7f;

	if ((*p & 0x7f) == 'h')
	{
		switch ((int) rval)
		{
		case 't': 
			rval = TH;
			p++;
			break;
		case 's': 
			rval = SH;
			p++;
			break;
		case 'p': 
			rval = PH;
			p++;
			break;
		case 'c': 
			rval = CH;
			p++;
			break;
		case 'g': 
			rval = GH;
			p++;
			break;
		}
	}

	*pp = p;
	return (rval);
}





/*------------------------------*/
/*	isweird			*/
/*------------------------------*/
static int isweird (x, y, p)
char	x,
	y,
       *p;
{

/*
 *	Return true if the string pointed to by p ends in:
 *		XYer XYers XYage XYages XYest
 *
 *	 where XY is one of the pairs:
 *		ft ld mp nd ng ns nt rg rm rn rt st
 */

	register unsigned int   c1,
	                        c2,
	                        c3;

	c1 = *p++ & 0x7f;
	c2 = *p++ & 0x7f;
	c3 = *p   & 0x7f;

	x &= 0x7f;
	y &= 0x7f;

	return (
			(
			    (c1 == 'e' && c2 == 'r')
			    || (c1 == 'a' && c2 == 'g' && c3 == 'e')
			    || (c1 == 'e' && c2 == 's' && c3 == 't')
			)
			&&
			(
			    (x == 'f' && y == 't')
			    || (x == 'l' && y == 'd')
			    || (x == 'm' && y == 'p')
			    || (x == 's' && y == 't')
			    || (x == 'n' && index ("dgst", (int) (y & 0x7f)))
			    || (x == 'r' && index ("gmnt", (int) (y & 0x7f)))
			)
		);
}




/*------------------------------*/
/*	consonants		*/
/*------------------------------*/
static void consonants (beg, end)
char   *beg,
       *end;
{

/*
 *	Hyphenate consonant pairs.
 *
 *	Look for a VCC pattern: use the state machine:
 *
 *		 V	    C	     C 
 * Start ----> 1 -----> 2 -----> 3 -----> 4
 *	   ^   |C   ^	| V	 | V	  | (no fetch )
 *	   |   |    |	v	 |	  |
 *	   |<--+    +------<-----+	  |
 *	   |				  |
 *	   +------------------------------+
 *
 *	The machine is implemented explicitly (with whiles and
 *	ifs and such) rather than with a table driven engine.
 *	A new character is fetched in states 1, 2, & 3 (but not 4).
 *	When you exit reach state 4 c1 and c2 will hold the two
 *	consonants, cp will point to the beginning of the second
 *	consonant, beg will point just past the second consonant.
 */

	extern char    *vccv_except[];

	register char	c1,
		        c2,
	               *cp;
	register char  *p;



	while (1)
	{
state1: 
		do
		{
			c2 = nextch (&beg, end);

		} while (isconsonant (c2));

		do
		{
			for (c1 = c2; isvowel (c1);)
			{
				cp = beg;
				c1 = nextch (&beg, end);
			}

			if (c1 == 'q' && *beg == 'u')	/* 	   Vqu */
			{
				HYPHENATE (*cp);	/* 	   V-qu */
				nextch (&beg, end);	/*   skip the u */
				goto state1;
			}

			cp = beg;
			c2 = nextch (&beg, end);

		} while (isvowel (c2));

		if (!c1 || !c2)
			break;

		/*
		 *   At this point we have found a Vowel-Cons-Cons
		 *   sequence. C1 and c2 will hold the left and right
		 *   consonant respectively. Beg will point at the
		 *   character following the second consonant. Cp
		 *   will point at the second consant.
		 */

		if (c1 == 'c' && c2 == 'k')	/* 	  Vck  */
		{
			if (*beg)
				HYPHENATE (*beg);/* 	  Vck- */
		}
		else if (c1 == c2)		/*     ller(s) */
		{
			if ((c1 != 'l' && c1 != 's')
			|| (isvowel (*beg) && !ER (beg, end)))
				HYPHENATE (*cp);
		}
		else if (isvowel (*beg))	/* 	  VCCV */
		{
			if (!isweird (c1, c2, beg))
			{
				if (!((p = vccv_except[(int) c1 - 'a'])
				&& index (p, (int) c2)))
					HYPHENATE (*cp);
			}
		}
	}

	UNHYPHENATE (*end);
}




/*------------------------------*/
/*	phyphen			*/
/*------------------------------*/
static void phyphen (start, end)
char   *start,
       *end;
{

/*
 *	Print a word with all the hyphens visible
 */

	printf ("Possible hyphenation points: <");

	for (; *start && start <= end; start++)
	{
		if (HAS_HYPHEN (*start))	/* put hyphen BEFORE char */
			putchar ('-');
		putchar ((int) (*start & 0x7f));
	}

	printf (">\n");
}




/*------------------------------*/
/*	next			*/
/*------------------------------*/
static int next (cur_state, cur_char)
char	cur_char;
int	cur_state;
{

/*
 *	Given the current state and the current input character
 *	return the next state. The global variable States must
 *	point at the correct state table (Suffixes or Prefixes).
 */

	extern char    *Prefixes[];	/* Prefix state table */
	extern char   **States;		/* Ptr to table for current state machine*/

	char	       *p = States[cur_state];
	int   		rval,
	       		i,
		        c;



	c = cur_char & 0x7f;

#ifdef DEBUG_HYPHEN
	if (HyDebug)
		printf ("%s: Current state = %d, Input char = <%c>, ",
				States == Prefixes ? "PREFIX" : "SUFFIX",
				cur_state, c);
#endif

	if (!*p)
		rval = (int) (p[(c - 'a') + 1]);
	else
	{
		for (rval = 0, i = *p++; --i >= 0; p += 2)
		{
			if (c == p[0])
			{
				rval = p[1];
				break;
			}
		}
	}

#ifdef DEBUG_HYPHEN
	if (HyDebug)
		printf ("Next state = %d\n", rval);
#endif

	return (rval);
}


/*
 *	State machine tables used by hyphen.c:
 *
 *	There are two types of states. If the first entry is 0 then the state
 *	is an array holding the next state, indexed by the current input
 *	character. If the first entry is non-zero then the table is a list of
 *	char/next-state pairs. The first byte is a character and the second is
 *	the state to go to on that character. The leftmost byte in the list is
 *	the number of entries in the list. All unspecified transitions are to
 *	zero.
 *
 * ----------------------------------------------------------------------
 *
 *	S(x) is used to initialize the state tables. It turned out to
 *	be easier to use a define than a typedef.
 */

#define SUC static char

SUC s0[] ={ 0, 84,84,84,1,1,84,84,84,84,84,84,84,84,84,84,84,84,84,1,84,
						 84,84,84,84,84,84 };
SUC s1[] ={ 0, 0,0,2,0,7,0,21,0,0,0,0,23,0,30,0,0,0,33,38,45,0,0,0,0,49,0};
SUC s2[] ={ 1, 'i', 3  };
SUC s3[] ={ 1, 'p', 4  };
SUC s4[] ={ 1, 'o', 5  };
SUC s5[] ={ 1, 'c', 6  };
SUC s6[] ={ 1, 's', 87 };
SUC s7[] ={ 0, 0,0,0,0,0,0,0,0,0,0,0,8,0,0,0,4,0,10,0,16,0,15,0,0,0,19 };
SUC s8[] ={ 1, 'b', 9  };
SUC s9[] ={ 1, 'a', 55 };
SUC s10[]={ 2, 'e', 11, 'u', 14 };
SUC s11[]={ 1, 'h', 12 };
SUC s12[]={ 1, 'p', 6  };
SUC s13[]={ 2, 'n', 81, 'r', 81 };
SUC s14[]={ 1, 't', 87 };
SUC s15[]={ 1, 'i', 14 };
SUC s16[]={ 1, 'a', 17 };
SUC s17[]={ 2, 'c', 18, 'l', 18 };
SUC s18[]={ 0, 88,0,0,0,88,0,0,0,88,0,0,0,0,0,88,0,0,0,0,0,88,0,0,0,88,0};
SUC s19[]={ 1, 'i', 20 };
SUC s20[]={ 1, 'l', 80 };
SUC s21[]={ 1, 'n', 22 };
SUC s22[]={ 1, 'i', 86 };
SUC s23[]={ 2, 'a', 25, 'u', 24 };
SUC s24[]={ 1, 'f', 83 };
SUC s25[]={ 0, 0,0,87,0,0,0,0,0,26,0,0,0,0,27,0,0,0,0,0,87,0,0,0,0,0,0 };
SUC s26[]={ 2, 't', 87, 'c', 87 };
SUC s27[]={ 1, 'o', 28 };
SUC s28[]={ 1, 'i', 29 };
SUC s29[]={ 1, 't', 89 };
SUC s30[]={ 1, 'o', 31 };
SUC s31[]={ 1, 'i', 32 };
SUC s32[]={ 2, 't', 87, 'c', 6 };
SUC s33[]={ 1, 'e', 34 };
SUC s34[]={ 1, 'h', 35 };
SUC s35[]={ 1, 'p', 36 };
SUC s36[]={ 1, 'a', 37 };
SUC s37[]={ 1, 'r', 87 };
SUC s38[]={ 2, 's', 39, 'u', 41 };
SUC s39[]={ 1, 'e', 40 };
SUC s40[]={ 2, 'l', 83, 'n', 83 };
SUC s41[]={ 1, 'o', 42 };
SUC s42[]={ 1, 'i', 43 };
SUC s43[]={ 1, 'c', 44 };
SUC s44[]={ 0, 88,88,88,88,88,88,88,88,88,88,88,88,88,88,88,88,88,88,87,88,
						  88,88,88,88,88,88 };
SUC s45[]={ 1, 'n', 46 };
SUC s46[]={ 1, 'e', 47 };
SUC s47[]={ 0, 0,0,0,87,0,0,0,0,48,0,0,0,83,0,0,0,0,0,0,0,0,0,0,0,0,0 };
SUC s48[]={ 1, 'c', 87 };
SUC s49[]={ 0, 0,0,0,0,0,0,50,35,0,0,0,83,0,0,0,0,0,51,0,0,0,0,0,0,0,0 };
SUC s50[]={ 1, 'o', 87 };
SUC s51[]={ 1, 'a', 52 };
SUC s52[]={ 1, 'n', 53 };
SUC s53[]={ 0, 88,88,88,88,81,88,88,88,88,88,88,88,88,88,54,88,88,88,88,88,
						  88,88,88,88,88,88 };
SUC s54[]={ 1, 'i', 82 };
SUC s55[]={ 0, 0,0,0,0,80,0,0,80,80,0,80,80,0,0,80,0,0,0,0,13,80,80,80,80,
								80,0};
char  *Suffixes[] =
{
	s0, s1, s2, s3, s4, s5, s6, s7, s8, s9,
	s10, s11, s12, s13, s14, s15, s16, s17, s18, s19,
	s20, s21, s22, s23, s24, s25, s26, s27, s28, s29,
	s30, s31, s32, s33, s34, s35, s36, s37, s38, s39,
	s40, s41, s42, s43, s44, s45, s46, s47, s48, s49,
	s50, s51, s52, s53, s54, s55
};



SUC p0[] = {1, 0, 0 };
SUC p1[] = {0, 0,2,4,6,9,0,0,13,22,0,0,26,29,39,41,45,50,0,53,57,64,
							0,0,0,0,0 };
SUC p2[] = {1, 'e', 3 };
SUC p3[] = {0, 0,0,81,0,0,0,0,81,0,0,0,0,0,0,0,0,0,0,81,0,0,0,81,0,0,0 };
SUC p4[] = {1, 'o', 5 };
SUC p5[] = {2, 'm', 87, 'n', 87 };
SUC p6[] = {1, 'i', 7 };
SUC p7[] = {1, 's', 8 };
SUC p8[] = {0, 85,85,85,85,85,85,85, 0,85,85,85,85,85,85,85,85,85,85,85,85,
						  85,85,85,85, 0,85 };
SUC p9[] = {2, 'q', 10, 'x', 87 };
SUC p10[]= {1, 'u', 11 };
SUC p11[]= {1, 'i', 12 };
SUC p12[]= {0, 81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,81,
						  81, 0,81,81,81,81 };
SUC p13[]= {0, 14,0,0,0,0,0,0,0,0,0,0,0,0,0,16,0,0,0,0,0,0,0,0,0,19,0 };
SUC p14[]= {1, 'n', 15 };
SUC p15[]= {1, 'd', 87 };
SUC p16[]= {1, 'r', 17 };
SUC p17[]= {1, 's', 18 };
SUC p18[]= {1, 'e', 87 };
SUC p19[]= {1, 'p', 20 };
SUC p20[]= {1, 'e', 21 };
SUC p21[]= {1, 'r', 84 };
SUC p22[]= {2, 'n', 69, 'm', 86 };
SUC p23[]= {0, 81,0,0,0,0,81,81,0,0,0,0,81,81,0,0,0,0,0,0,0,0,0,0,0,0,0 };
SUC p24[]= {2, 'e', 21, 'r', 25 };
SUC p25[]= {1, 'o', 87 };
SUC p26[]= {1, 'e', 27 };
SUC p27[]= {1, 'x', 35 };
SUC p28[]= {1, 'i', 87 };
SUC p29[]= {0, 30,0,0,0,0,0,0,0,36,0,0,0,0,0,0,0,0,0,0,0,37,0,0,0,0,0 };
SUC p30[]= {0, 0,0,31,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33,0,0,0,35,0,0 };
SUC p31[]= {1, 'r', 32 };
SUC p32[]= {1, 'o', 83 };
SUC p33[]= {1, 'h', 34 };
SUC p34[]= {1, 'e', 82 };
SUC p35[]= {1, 'i', 82 };
SUC p36[]= {1, 'n', 35 };
SUC p37[]= {1, 'l', 38 };
SUC p38[]= {1, 't', 65 };
SUC p39[]= {1, 'o', 40 };
SUC p40[]= {1, 'n', 86 };
SUC p41[]= {2, 'u', 42, 'v', 43 };
SUC p42[]= {1, 't', 87 };
SUC p43[]= {1, 'e', 44 };
SUC p44[]= {1, 'r', 86 };
SUC p45[]= {1, 's', 46 };
SUC p46[]= {1, 'e', 47 };
SUC p47[]= {1, 'u', 48 };
SUC p48[]= {1, 'd', 49 };
SUC p49[]= {1, 'o', 83 };
SUC p50[]= {1, 'u', 51 };
SUC p51[]= {1, 'a', 52 };
SUC p52[]= {1, 'd', 87 };
SUC p53[]= {0, 0,0,0,0,54,0,0,0,0,0,0,0,0,0,55,0,0,0,0,0,56,0,0,0,0,0 };
SUC p54[]= {1, 'm', 28 };
SUC p55[]= {1, 'm', 18 };
SUC p56[]= {2, 'b', 87, 'p', 20 };
SUC p57[]= {2, 'h', 58, 'r', 61 };
SUC p58[]= {1, 'e', 59 };
SUC p59[]= {1, 'r', 60 };
SUC p60[]= {1, 'e', 87 };
SUC p61[]= {2, 'a', 62, 'i', 68 };
SUC p62[]= {1, 'n', 63 };
SUC p63[]= {1, 's', 23 };
SUC p64[]= {1, 'n', 66 };
SUC p65[]= {1, 'i', 83 };
SUC p66[]= {0, 85,85,85,70,85,85,85,85, 0,85,85,85,85,85,85,85,85,85,
					85,85,85,85,85,85,85,85 };
SUC p67[]= {1, 'r', 87 };
SUC p68[]= {0, 81,0,0,0,0,81,0,0,0,0,0,0,0,0,0,0,0,0,0,0,81,0,0,0,0,0 };
SUC p69[]= {0, 85,85,85,85,85,85,85,85,85,85,85,85,85,85,85,85,85,85,
					85,24,85,85,85,85,85,85 };
SUC p70[]= {1, 'e', 67 };


char  *Prefixes[] =
{
	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9,
	p10, p11, p12, p13, p14, p15, p16, p17, p18, p19,
	p20, p21, p22, p23, p24, p25, p26, p27, p28, p29,
	p30, p31, p32, p33, p34, p35, p36, p37, p38, p39,
	p40, p41, p42, p43, p44, p45, p46, p47, p48, p49,
	p50, p51, p52, p53, p54, p55, p56, p57, p58, p59,
	p60, p61, p62, p63, p64, p65, p66, p67, p68, p69,
	p70
};





/*
 *	except - exception processing for the hyphenation alg
 *
 *	All of the following words are exceptions to the hyphenation
 *	rules. The words in this list will also be recognized if the
 *	are followed by an s or es.
 */
static char *Hyexcept[] =
{
"ab-sorb-ent",		"ac-cept-able",		"ac-ceptor",
"ac-cord-ance",		"ac-count-ant",		"ac-know-ledge",
"adapt-able",		"adapt-er",		"af-firm-ative",
"al-go-rithm",		"an-nouncer",		"ant-acid",
"ant-eater",		"ant-hill",		"an-tiq-uity",
"any-thing",		"apart-heid",		"arch-an-gel",
"arch-ery",		"ar-mi-stice",		"art-ist",
"askance",		"astig-ma-tism",	"astir",
"astonish-ment",	"at-mos-phere",		"bal-ding",
"bar-on-ess",		"beach-comber",		"beck-on",
"bes-tial",		"be-tween",		"bib-li-og-raphy",
"bind-ery",		"bi-no-mial",		"blast-off",
"board-er",		"bomb-er",		"bouncer",
"bound-ary",		"buff-er",		"bull-ish",
"buzz-er",		"by-stand-er",		"candle-stick",
"carb-on",		"cast-away",		"cast-off",
"cat-ion",		"cav-ern-ous",		"cen-ter",
"change-over",		"child-ish",		"chordal",
"civ-i-lize",		"class-ify",		"class-room",
"climb-er",		"clinch-er",
"coars-en",		"cognac",
"cole-slaw",		"com-a-tose",		"com-bat-ive",
"come-back",		"co-me-dian",		"com-men-da-tory",
"comp-troller",		"com-put-abil-ity",	"con-de-scend",
"cone-flower",		"con-form-ity",		"con-sult-ant",
"con-test-ant",		"con-trol-lable",	"co-nun-drum",
"con-vert-ible",	"co-star",		"count-ess",
"court-house",		"court-yard",		"cre-scendo",
"crest-fallen",		"cross-over",		"crypt-analysis",
"crys-tal-lize",	"curl-i-cue",		"damp-en",
"damp-est",		"dar-ling",		"debt-or",
"dec-la-ra-tion",	"de-cre-scendo",	"de-duct-ible",
"de-form-ity",		"de-gree",		"de-pend-able",
"de-pend-ent",		"de-scend-ent",		"de-scent",
"de-test-able",		"di-gest-ible",		"dis-cern-ible",
"dis-miss-al",		"dis-till-ery",		"dump-ling",
"earth-en-ware",	"east-ern",		"egg-head",
"egg-nog",		"eld-est",		"else-where",
"eq-uable",		"equipped",		"err-ing",
"es-tab-lish",		"eu-logy",		"eve-ning",
"every-thing",		"ex-ac-ti-tude",	"ex-ist-ence",
"ex-pend-able",		"ex-press-ible",	"fall-out",
"false-hood",		"far-thing",		"fencer",
"fiend-ish",		"for-eign-er",		"fore-short-en",
"fore-stall",		"found-ling",		"gen-er-ator",
"gold-en",		"handle-bar",		"hang-out",
"hang-over",		"hap-hazard",		"ha-rangue",
"hard-en",		"hard-ened",		"hard-est",
"harp-ist",		"haz-ard-ous",		"heart-ache",
"heart-ily",		"hence-forth",		"her-bal",
"hogs-head",		"hold-out",		"hold-over",
"hold-up",
"idler",		"im-mo-bi-lize",
"im-pass-able",		"im-per-turb-able",	"inch-worm",
"in-clem-ent",		"in-con-test-able",	"in-de-pend-ent",
"in-di-gest-ible",	"ineq-uity",		"in-ex-acti-tude",
"in-ex-haust-ible",	"in-form-ant",		"iniq-uity",
"ink-blot",		"ink-ling",		"inn-keeper",
"in-sa-tiable",		"in-te-rior",		"in-ter-rupt-ible",
"ir-re-vers-ible",	"i-so-ther-mal",	"jeop-ard-ize",
"kib-itzer",
"land-owner",		"launch-er",		"left-ist",
"left-over",		"less-en",		"life-style",
"lift-off",		"lime-stone",		"li-on-ess",
"liq-uefy",		"liq-uid",		"liq-ui-date",
"liq-ui-da-tion",	"liq-uor",		"live-stock",
"lull-aby",		"lunch-eon",		"lus-cious",
"main-spring",		"mast-head",		"me-ringue",
"me-ta-bo-lize",	"met-al",		"mile-stone",
"mince-meat",		"min-is-ter",		"min-is-try",
"mo-bi-lize",		"mod-ern-ize",		"mo-nop-o-lize",
"morgue",		"needle-work",		"neg-li-gible",
"ne-go-tiable",		"nerv-ous",		"nest-ling",
"non-con-form-ist",	"none-the-less",	"non-ex-ist-ent",
"non-metal",		"north-east",		"north-ern",
"nurse-maid",		"nurs-ery",		"ob-serv-able",
"ob-server",		"off-beat",		"off-hand",
"off-print",		"off-shoot",		"off-shore",
"off-spring",		"orange-ade",		"out-land-ish",
"pal-ate",		"pass-able",		"ped-a-gogy",
"pent-house",		"per-cent-age",		"pe-ri-odic",
"per-sist-ent",		"pet-al",		"pho-to-stat",
"play-thing",		"pleb-i-scite",		"plumb-er",
"poly-no-mial",		"port-hole",		"post-al",
"post-hu-mous",		"pre-dict-able",	"pre-req-ui-site",
"pre-school",		"pre-serv-ative",	"pre-vious",
"priest-hood",		"prob-abil-ity",	"prob-able",
"pro-ce-dure",		"pro-gram",		"pro-gram-mer",
"pro-grams",		"psalm-ist",		"pub-li-ca-tion",
"pub-lish",		"qua-drille",		"ranch-er",
"rattle-snake",		"re-corder",		"re-hears-al",
"rent-al",		"re-place-ment",	"rep-re-sentative",
"req-ui-si-tion",	"re-scind",		"re-search-er",
"re-solv-able",		"re-spect-able",	"re-start-ed",
"re-state-ment",	"re-store",		"re-vers-ible",
"re-volv-er",		"roll-away",		"round-about",
"sap-ling",		"sea-scape",		"self-ish",
"sell-out",		"send-off",		"sense-less",
"serv-er",		"serv-ice-able",	"sharpen",
"shoe-string",		"short-en",		"shy-ster",
"sib-ling",		"side-step",		"side-swipe",
"si-lencer",	
"smoke-stack",
"snake-skin",		"so-ciable",		"soft-hearted",
"solv-able",		"som-er-sault",		"some-thing",
"sta-bi-lize",		"stand-ard-ize",	"stand-out",
"star-ling",		"stat-ure",		"ster-ling",
"stew-ard-ess",		"stiff-en",		"sub-se-quence",
"sug-gest-ible",	"su-pe-rior",		"surf-er",
"tan-ta-lize",		"thermo-stat",		"tongue",
"torque",		"toss-up",		"trench-ant",
"turn-about",		"turn-over",		"turn-table",
"ubiq-ui-tous",		"una-nim-ity",		"u-nan-i-mous",
"un-civ-i-lized",	"un-class-ified",	"un-con-trollable",
"unc-tuous",		"un-der-stand-able",	"un-err-ing",
"un-gov-ern-able",	"un-pre-dict-able",	"un-search-able",
"un-so-ciable",		"un-solv-able",		"up-swing",
"venge-ance",		"vict-ual",		"vignette",
"volt-age",		"wall-eye",		"waste-bas-ket",
"waste-land",		"watt-meter",		"weak-ling",
"west-ern-ize",		"when-ever",		"whisk-er",
"wors-en",		"yard-age",		"year-ling"
};

#define TABSIZE		(sizeof(Hyexcept)/sizeof(char *))


/*------------------------------*/
/*	exception		*/
/*------------------------------*/
static int exception (word, end)
char   *word,
       *end;
{

/*
 *	Looks up word in the exception list and returns TRUE
 *	if it's there (FALSE if it isn't).
 */

	char  **pp,
	       *p;
	char	oend;
	int	rval = 0;

	oend = *++end;
	*end = 0;

	/*
	 *   seach for word in exception list...
	 */
	pp = (char **) bsearch (word,
				 Hyexcept,
				 (unsigned int) TABSIZE,
				 (unsigned int) sizeof(char*),
				 wcmp);

	/*
	 *   if we found a match...
	 */
	if (pp)
	{
		/*
		 *   ..."hyphenate" the word (mark hyphen bit)
		 */
		for (p = *pp ; *p ; word++, p++)
		{
			if (*p == '-')
			{
				HYPHENATE (*word);
				p++;
			}
		}
		rval = 1;
	}

	*end = oend;

	return (rval);
}



/*------------------------------*/
/*	wcmp			*/
/*------------------------------*/
static int wcmp (word, tabp)
char   *word;
char  **tabp;
{

/*
 *	handles plurals. TRUE is returned if the input word matches (or
 *	matches and is followed by an 's' or 'es')
 */

	register char *ptab = *tabp;

	while (*word && *ptab)
	{
		/*
		 *   skip hyphen in table...
		 */
		if (*ptab == '-')
			ptab++;

		/*
		 *   stop at the first differnce...
		 */
		if (*word != *ptab)
			break;

		word++;
		ptab++;
	}

	/*
	 *   if at the end of the table, and input word is done, except for
	 *   plural ending, we matched...
	 */
	if (!*ptab && (word[0]=='s' || (word[0]=='e' && word[1]=='s')))
		return (0);

	/*
	 *   otherwise, return difference in char where match failed...
	 */
	return ((int) (*word - *ptab));
}




/*------------------------------*/
/*	bsearch			*/
/*------------------------------*/
static char *bsearch (key, srch, nsrch, size, compare)
char	       *key,		/* pointer to element to be matched */
	       *srch;		/* first element of the array to be searched */
unsigned int	nsrch,		/* number of members in the array */
		size;		/* num bytes each array element takes up */
int		(*compare)();	/* ptr to func that compares elts of array*/
{

/*
 *	bsearch - binary search 
 * 
 *	uses the standard binary search algorithm. the only difference
 *	is the addition of the two intial checks to see if the key value
 *	is in between the first and last element of the input array.  
 */

	char   *last_elt = srch + (size * (nsrch - 1));
	char   *middle_elt;
	int	compval;

	/*
	 *   quick check: is first elt >= key? if so, return...
	 */
	if ((*compare)(key, srch) < 0)
		return ((char *) NULL);

	/*
	 *   quick check: is last elt <= key? if so, return...
	 */
	if ((*compare)(key, last_elt) > 0)
		return ((char *) NULL);

	/*
	 *   ok, key falls between first and last. on with the show...
	 */
	while (last_elt >= srch)
	{
		/*
		 *   starts with full range and bisects set each iteration
		 *
		 *   find center elt...
		 */
		middle_elt = (char *) (srch + (size * (((int) last_elt -
				   (int) srch) / (2 *size))));
		/*
		 *   compare key with this...
		 */
		compval = (*compare)(key, middle_elt);

		/*
		 *   what happened?
		 */
		if (!compval)
			/* a match!*/
			return ((char *) middle_elt);
		else if (compval < 0)
			/* reset last elt... */
			last_elt = middle_elt - size;
		else
			/* reset first elt... */
			srch = middle_elt + size;
	}
	return ((char *) NULL);
}










/*
 *
 *	     MISC. DEBUGGING ROUTINES (INCLUDING A MAIN MODULE):
 */


#ifdef DEBUG_HYPHEN

int     Verbose = 0;

/*------------------------------*/
/*	bprint			*/
/*------------------------------*/
static void bprint (b, e)
char   *b,
       *e;
{

/*
 *	Print all characters between b and e inclusive
 */

	while (b <= e && *b)
		putchar (*b++ & 0x7f);
}

#endif /*DEBUG_HYPHEN*/









#ifdef STANDALONE


#ifdef DEBUG_HYPHEN
int     Verbose = 0;
#endif

/*------------------------------*/
/*	check_hyphen		*/
/*------------------------------*/
int check_hyphen (ps, ph)
register char  *ps;
register char **ph;
{
	static char	buf[132],
		       *pbuf;
	register int	c;

	*ph   = buf;
	pbuf  = buf;
	while (c = (int) *ps++)
		*pbuf++ = (char) (c & 0x7f);
	*pbuf-- = 0;

	return (hyphen (buf, pbuf));
}



/*------------------------------*/
/*	main			*/
/*------------------------------*/
main (argc, argv)
int     argc;
char   *argv[];
{
	int	ret;
	char   *phyph;
	char   *ps;

	argc--, argv++;

	while (argc && **argv == '-')
	{
		switch (*(*argv+1))
		{
		case 'd':
			HyDebug = 1;
			break;

		case 'v':
			Verbose = 1;
			break;
		}
		argc--, argv++;
	}
	if (HyDebug)
		Verbose = 2;

	while (argc--)
	{
		if (Verbose)
			printf ("check word: <%s>\n", *argv);

		ret = check_hyphen (*argv, &phyph);

		if (ret)
		{
			if (Verbose)
				printf ("\ncan hyphenate %s: ", *argv);
			ps = phyph;
			while (*ps)
			{
				if (HAS_HYPHEN (*ps))
					putchar ('-');
				putchar ((int) (*ps++ & 0x7f));
			}
			printf ("\n");
		}
		argv++;
	}
	exit (0);
}

#endif /*STANDALONE*/

